// 实现简单的SPN密码，用作例子
// 依赖libsodium
#pragma once
#ifndef SSPN_H_
#define SSPN_H_

#include <stdio.h>
#include <stdint.h>
#include <sodium.h>

#ifdef cplusplus
extern "C" {
#endif

#define BIT4    uint8_t
#define BIT16   uint16_t

BIT4 sbox(BIT4 in) {
    switch (in & 0x0f) {
        case 0x0:   return 0xe;
        case 0x1:   return 0x4;
        case 0x2:   return 0xd;
        case 0x3:   return 0x1;
        case 0x4:   return 0x2;
        case 0x5:   return 0xf;
        case 0x6:   return 0xb;
        case 0x7:   return 0x8;
        case 0x8:   return 0x3;
        case 0x9:   return 0xa;
        case 0xa:   return 0x6;
        case 0xb:   return 0xc;
        case 0xc:   return 0x5;
        case 0xd:   return 0x9;
        case 0xe:   return 0x0;
        case 0xf:   return 0x7;
    }
    return 0x0;
}

BIT4 sbox_inverse(BIT4 in) {
        switch (in & 0x0f) {
        case 0xe:   return 0x0;
        case 0x4:   return 0x1;
        case 0xd:   return 0x2;
        case 0x1:   return 0x3;
        case 0x2:   return 0x4;
        case 0xf:   return 0x5;
        case 0xb:   return 0x6;
        case 0x8:   return 0x7;
        case 0x3:   return 0x8;
        case 0xa:   return 0x9;
        case 0x6:   return 0xa;
        case 0xc:   return 0xb;
        case 0x5:   return 0xc;
        case 0x9:   return 0xd;
        case 0x0:   return 0xe;
        case 0x7:   return 0xf;
    }
    return 0x0;
}

BIT16 permutation(BIT16 in) {
    BIT16 result = 0x0000;
    result = result | (in & 0x8000) | (in & 0x0800) << 3 | (in & 0x0080) << 6 | (in & 0x0008) << 9;
    result = result | (in & 0x4000) >> 3 | (in & 0x0400) | (in & 0x0040) << 3 | (in & 0x0004) << 6;
    result = result | (in & 0x2000) >> 6 | (in & 0x0200) >> 3 | (in & 0x0020) | (in & 0x0002) << 3;
    result = result | (in & 0x1000) >> 9 | (in & 0x0100) >> 6 | (in & 0x0010) >> 3 | (in & 0x0001);
    return result;
}

BIT16 random16() {
    if (sodium_init() < 0) {
        return 0;
    }
    return (BIT16)randombytes_uniform(0x10000);
}

void subkey(BIT16 rk[5]) {
    if (sodium_init() < 0) {
        return;
    }
    for (int i = 0; i < 5; i++) {
        rk[i] = random16();
    }
}

void split(BIT16 in, BIT4 *b1, BIT4 *b2, BIT4 *b3, BIT4 *b4) {
    *b1 = (in >> 12) & 0x000f;
    *b2 = (in >>  8) & 0x000f;
    *b3 = (in >>  4) & 0x000f;
    *b4 = (in >>  0) & 0x000f;
}

BIT16 concat(BIT4 b1, BIT4 b2, BIT4 b3, BIT4 b4) {
    BIT16 result = 0x0000;
    result |= (BIT16)b1 << 12;
    result |= (BIT16)b2 << 8;
    result |= (BIT16)b3 << 4;
    result |= (BIT16)b4;
    return result;
}

BIT16 sbox16(BIT16 in) {
    BIT4 si1, si2, si3, si4, so1, so2, so3, so4;
    split(in, &si1, &si2, &si3, &si4);
    so1 = sbox(si1);
    so2 = sbox(si2);
    so3 = sbox(si3);
    so4 = sbox(si4);
    return concat(so1, so2, so3, so4);
}

BIT16 sbox16_inverse(BIT16 in) {
    BIT4 si1, si2, si3, si4, so1, so2, so3, so4;
    split(in, &si1, &si2, &si3, &si4);
    so1 = sbox_inverse(si1);
    so2 = sbox_inverse(si2);
    so3 = sbox_inverse(si3);
    so4 = sbox_inverse(si4);
    return concat(so1, so2, so3, so4);
}

BIT16 round(BIT16 in, BIT16 rk) {
    // key mixing
    in = in ^ rk;
    // sbox
    BIT16 so = sbox16(in);
    // permutation
    return permutation(so);
}

BIT16 round_inverse(BIT16 in, BIT16 rk) {
    // permutation
    BIT16 so = permutation(in);
    // sbox inserver
    BIT16 si = sbox16_inverse(so);
    // key mixing
    return si ^ rk;
}

BIT16 encrypt(BIT16 plain, BIT16 rk[5]) {
    BIT16 m1 = round(plain, rk[0]);
    BIT16 m2 = round(m1, rk[1]);
    BIT16 m3 = round(m2, rk[2]);
    BIT16 m4 = round(m3, rk[3]);
    return permutation(m4) ^ rk[4];
}

BIT16 decrypt(BIT16 cipher, BIT16 rk[5]) {
    BIT16 m0 = permutation(cipher ^ rk[4]);
    BIT16 m1 = round_inverse(m0, rk[3]);
    BIT16 m2 = round_inverse(m1, rk[2]);
    BIT16 m3 = round_inverse(m2, rk[1]);
    BIT16 m4 = round_inverse(m3, rk[0]);
    return m4;
}

#ifdef cplusplus
}
#endif

#endif